@@ -3,7 +3,7 @@ package PDF::API2;
 use strict;
 no warnings qw[ deprecated recursion uninitialized ];
 
-our $VERSION = '2.036'; # VERSION
+our $VERSION = '2.041'; # VERSION
 
 use Carp;
 use Encode qw(:all);
@@ -47,7 +47,7 @@ PDF::API2 - Facilitates the creation and modification of PDF files
     $page = $pdf->page();
 
     # Retrieve an existing page
-    $page = $pdf->openpage($page_number);
+    $page = $pdf->open_page($page_number);
 
     # Set the page size
     $page->mediabox('Letter');
@@ -73,9 +73,9 @@ PDF::API2 - Facilitates the creation and modification of PDF files
 
 =item $pdf = PDF::API2->new(%options)
 
-Creates a new PDF object.  If you will be saving it as a file and
-already know the filename, you can give the '-file' option to minimize
-possible memory requirements later on.
+Creates a new PDF object.  If you will be saving it as a file and already know
+the filename, you can give the '-file' option to minimize possible memory
+requirements later on.
 
 B<Example:>
 
@@ -91,6 +91,9 @@ B<Example:>
     ...
     $pdf->save();
 
+To turn off automatic compression of PDF contents, include option C<-compress>
+with a false value.  This is generally only useful for debugging.
+
 =cut
 
 sub new {
@@ -100,7 +103,7 @@ sub new {
     bless $self, $class;
     $self->{'pdf'} = PDF::API2::Basic::PDF::File->new();
 
-    $self->{'pdf'}->{' version'} = 4;
+    $self->{'pdf'}->{' version'} = '1.4';
     $self->{'pages'} = PDF::API2::Basic::PDF::Pages->new($self->{'pdf'});
     $self->{'pages'}->proc_set(qw(PDF Text ImageB ImageC ImageI));
     $self->{'pages'}->{'Resources'} ||= PDFDict();
@@ -109,7 +112,12 @@ sub new {
     weaken $self->{'catalog'};
     $self->{'fonts'} = {};
     $self->{'pagestack'} = [];
-    $self->{'forcecompress'} = 1;
+    if (exists $options{'-compress'}) {
+        $self->{'forcecompress'} = $options{'-compress'} ? 1 : 0;
+    }
+    else {
+        $self->{'forcecompress'} = 1;
+    }
     $self->preferences(%options);
     if ($options{'-file'}) {
         $self->{'pdf'}->create_file($options{'-file'});
@@ -123,7 +131,7 @@ sub new {
     return $self;
 }
 
-=item $pdf = PDF::API2->open($pdf_file)
+=item $pdf = PDF::API2->open($pdf_file, %options)
 
 Opens an existing PDF file.
 
@@ -137,6 +145,9 @@ B<Example:>
     ...
     $pdf->update();
 
+To turn off automatic compression of PDF contents, include option C<-compress>
+with a false value.  This is generally only useful for debugging.
+
 =cut
 
 sub open {
@@ -144,30 +155,49 @@ sub open {
     croak "File '$file' does not exist" unless -f $file;
     croak "File '$file' is not readable" unless -r $file;
 
-    my $content;
-    my $scalar_fh = FileHandle->new();
-    CORE::open($scalar_fh, '+<', \$content) or die "Can't begin scalar IO";
-    binmode $scalar_fh, ':raw';
-
-    my $disk_fh = FileHandle->new();
-    CORE::open($disk_fh, '<', $file) or die "Can't open $file for reading: $!";
-    binmode $disk_fh, ':raw';
-    $disk_fh->seek(0, 0);
-    my $data;
-    while (not $disk_fh->eof()) {
-        $disk_fh->read($data, 512);
-        $scalar_fh->print($data);
+    my $self = {};
+    bless $self, $class;
+    foreach my $parameter (keys %options) {
+        $self->default($parameter, $options{$parameter});
     }
-    $disk_fh->close();
-    $scalar_fh->seek(0, 0);
 
-    my $self = $class->open_scalar($content, %options);
+    my $is_writable = -w $file;
+    $self->{'pdf'} = PDF::API2::Basic::PDF::File->open($file, $is_writable);
+    _open_common($self, %options);
     $self->{'pdf'}->{' fname'} = $file;
+    $self->{'opened_readonly'} = 1 unless $is_writable;
 
     return $self;
 }
 
-=item $pdf = PDF::API2->open_scalar($pdf_string)
+sub _open_common {
+    my ($self, %options) = @_;
+
+    $self->{'pdf'}->{'Root'}->realise();
+    $self->{'pdf'}->{' version'} ||= '1.3';
+
+    $self->{'pages'} = $self->{'pdf'}->{'Root'}->{'Pages'}->realise();
+    weaken $self->{'pages'};
+    my @pages = proc_pages($self->{'pdf'}, $self->{'pages'});
+    $self->{'pagestack'} = [sort { $a->{' pnum'} <=> $b->{' pnum'} } @pages];
+    weaken $self->{'pagestack'}->[$_] for (0 .. scalar @{$self->{'pagestack'}});
+
+    $self->{'catalog'} = $self->{'pdf'}->{'Root'};
+    weaken $self->{'catalog'};
+
+    $self->{'opened'} = 1;
+    if (exists $options{'-compress'}) {
+        $self->{'forcecompress'} = $options{'-compress'} ? 1 : 0;
+    }
+    else {
+        $self->{'forcecompress'} = 1;
+    }
+    $self->{'fonts'} = {};
+    $self->{'infoMeta'} = [qw(Author CreationDate ModDate Creator Producer Title Subject Keywords)];
+    return $self;
+}
+
+=item $pdf = PDF::API2->open_scalar($pdf_string, %options)
 
 Opens a PDF contained in a string.
 
@@ -182,6 +212,9 @@ B<Example:>
     ...
     $pdf->saveas('our/new.pdf');
 
+To turn off automatic compression of PDF contents, include option C<-compress>
+with a false value.  This is generally only useful for debugging.
+
 =cut
 
 # Deprecated (renamed)
@@ -201,19 +234,8 @@ sub open_scalar {
     CORE::open($fh, '+<', \$content) or die "Can't begin scalar IO";
 
     $self->{'pdf'} = PDF::API2::Basic::PDF::File->open($fh, 1);
-    $self->{'pdf'}->{'Root'}->realise();
-    $self->{'pages'} = $self->{'pdf'}->{'Root'}->{'Pages'}->realise();
-    weaken $self->{'pages'};
-    $self->{'pdf'}->{' version'} ||= 3;
-    my @pages = proc_pages($self->{'pdf'}, $self->{'pages'});
-    $self->{'pagestack'} = [sort { $a->{' pnum'} <=> $b->{' pnum'} } @pages];
-    weaken $self->{'pagestack'}->[$_] for (0 .. scalar @{$self->{'pagestack'}});
-    $self->{'catalog'} = $self->{'pdf'}->{'Root'};
-    weaken $self->{'catalog'};
+    _open_common($self, %options);
     $self->{'opened_scalar'} = 1;
-    $self->{'forcecompress'} = 1;
-    $self->{'fonts'} = {};
-    $self->{'infoMeta'} = [qw(Author CreationDate ModDate Creator Producer Title Subject Keywords)];
 
     return $self;
 }
@@ -582,10 +604,10 @@ sub version {
     my $self = shift();
     if (scalar @_) {
         my $version = shift();
-        croak "Invalid version $version" unless $version =~ /^(?:1\.)?([0-9]+)$/;
+        croak "Invalid version $version" unless $version =~ /^([12]\.[0-9]+)$/;
         $self->{'pdf'}->{' version'} = $1;
     }
-    return '1.' . $self->{'pdf'}->{' version'};
+    return $self->{'pdf'}->{' version'};
 }
 
 =item $bool = $pdf->isEncrypted()
@@ -938,7 +960,8 @@ B<Example:>
 
 sub update {
     my $self = shift();
-    $self->saveas($self->{'pdf'}->{' fname'});
+    croak "File is read-only" if $self->{'opened_readonly'};
+    $self->{'pdf'}->close_file();
     return;
 }
 
@@ -969,7 +992,16 @@ sub saveas {
         $self->{'pdf'}->close_file();
     }
     else {
-        $self->{'pdf'}->out_file($file);
+        unless ($self->{'pdf'}->{' fname'}) {
+            $self->{'pdf'}->out_file($file);
+        }
+        elsif ($self->{'pdf'}->{' fname'} eq $file) {
+            $self->update();
+        }
+        else {
+            $self->{'pdf'}->clone_file($file);
+            $self->{'pdf'}->close_file();
+        }
     }
 
     $self->end();
@@ -1023,6 +1055,13 @@ sub stringify {
         $self->{'pdf'}->append_file();
         $str = ${$self->{'content_ref'}};
     }
+    elsif ($self->{'opened'}) {
+        my $fh = FileHandle->new();
+        CORE::open($fh, '>', \$str) || die "Can't begin scalar IO";
+        $self->{'pdf'}->clone_file($fh);
+        $self->{'pdf'}->close_file();
+        $fh->close();
+    }
     else {
         my $fh = FileHandle->new();
         CORE::open($fh, '>', \$str) || die "Can't begin scalar IO";
@@ -1125,7 +1164,7 @@ sub page {
     return $page;
 }
 
-=item $page = $pdf->openpage($page_number)
+=item $page = $pdf->open_page($page_number)
 
 Returns the L<PDF::API2::Page> object of page $page_number.
 
@@ -1135,14 +1174,17 @@ document.
 B<Example:>
 
     $pdf = PDF::API2->open('our/99page.pdf');
-    $page = $pdf->openpage(1);   # returns the first page
-    $page = $pdf->openpage(99);  # returns the last page
-    $page = $pdf->openpage(-1);  # returns the last page
-    $page = $pdf->openpage(999); # returns undef
+    $page = $pdf->open_page(1);   # returns the first page
+    $page = $pdf->open_page(99);  # returns the last page
+    $page = $pdf->open_page(-1);  # returns the last page
+    $page = $pdf->open_page(999); # returns undef
 
 =cut
 
-sub openpage {
+# Deprecated (renamed)
+sub openpage { return open_page(@_); } ## no critic
+
+sub open_page {
     my $self = shift();
     my $index = shift() || 0;
     my ($page, $rotate, $media, $trans);
@@ -1165,7 +1207,7 @@ sub openpage {
         weaken $page->{' apipdf'};
         weaken $page->{' api'};
         $self->{'pdf'}->out_obj($page);
-        if (($rotate = $page->find_prop('Rotate')) and (not defined($page->{' fixed'}) or $page->{' fixed'} < 1)) {
+        if (($rotate = $page->find_prop('Rotate')) and not $page->{' opened'}) {
             $rotate = ($rotate->val() + 360) % 360;
 
             if ($rotate != 0 and not $self->default('nounrotate')) {
@@ -1200,7 +1242,7 @@ sub openpage {
             $trans = '';
         }
 
-        if (defined $page->{'Contents'} and (not defined($page->{' fixed'}) or $page->{' fixed'} < 1)) {
+        if (defined $page->{'Contents'} and not $page->{' opened'}) {
             $page->fixcontents();
             my $uncontent = delete $page->{'Contents'};
             my $content = $page->gfx();
@@ -1217,19 +1259,15 @@ sub openpage {
                 $content->{' stream'} .= ' Q ';
             }
 
-            ## $content->{'Length'} = PDFNum(length($content->{' stream'}));
-            # this will be fixed by the following code or content or filters
-
-            ## if we like compress we will do it now to do quicker saves
-            if ($self->{'forcecompress'} > 0) {
-                # $content->compressFlate();
+            # if we like compress we will do it now to do quicker saves
+            if ($self->{'forcecompress'}) {
                 $content->{' stream'} = dofilter($content->{'Filter'}, $content->{' stream'});
                 $content->{' nofilt'} = 1;
                 delete $content->{'-docompress'};
                 $content->{'Length'} = PDFNum(length($content->{' stream'}));
             }
         }
-        $page->{' fixed'} = 1;
+        $page->{' opened'} = 1;
     }
 
     $self->{'pdf'}->out_obj($page);
@@ -1335,7 +1373,7 @@ sub importPageIntoForm {
         $s_page = $s_idx;
     }
     else {
-        $s_page = $s_pdf->openpage($s_idx);
+        $s_page = $s_pdf->open_page($s_idx);
     }
 
     $self->{'apiimportcache'} ||= {};
@@ -1369,23 +1407,23 @@ sub importPageIntoForm {
 
     # create a whole content stream
     ## technically it is possible to submit an unfinished
-    ## (eg. newly created) source-page, but thats nonsense,
-    ## so we expect a page fixed by openpage and die otherwise
-    die "page not processed via openpage ..." unless $s_page->{' fixed'} == 1;
+    ## (eg. newly created) source-page, but that's nonsense,
+    ## so we expect a page fixed by open_page and die otherwise
+    unless ($s_page->{' opened'}) {
+        croak join(' ',
+                   "Pages may only be imported from a complete PDF.",
+                   "Save and reopen the source PDF object first");
+    }
 
-    # since the source page comes from openpage it may already
-    # contain the required starting 'q' without the final 'Q'
-    # if forcecompress is in effect
     if (defined $s_page->{'Contents'}) {
         $s_page->fixcontents();
 
         $xo->{' stream'} = '';
-        # openpage pages only contain one stream
+        # open_page pages only contain one stream
         my ($k) = $s_page->{'Contents'}->elements();
         $k->realise();
         if ($k->{' nofilt'}) {
-          # we have a finished stream here
-          # so we unfilter
+          # we have a finished stream here so we unfilter
           $xo->add('q', unfilter($k->{'Filter'}, $k->{' stream'}), 'Q');
         }
         else {
@@ -1393,7 +1431,7 @@ sub importPageIntoForm {
             # so we just copy it and add the required "qQ"
             $xo->add('q', $k->{' stream'}, 'Q');
         }
-        $xo->compressFlate() if $self->{'forcecompress'} > 0;
+        $xo->compressFlate() if $self->{'forcecompress'};
     }
 
     return $xo;
@@ -1442,7 +1480,7 @@ sub import_page {
         $s_page = $s_idx;
     }
     else {
-        $s_page = $s_pdf->openpage($s_idx);
+        $s_page = $s_pdf->open_page($s_idx);
     }
 
     if (ref($t_idx) eq 'PDF::API2::Page') {
@@ -2050,8 +2088,10 @@ Imports and returns a new PNM image object.  C<$file> may be either a filename o
 sub image_pnm {
     my ($self, $file, %opts) = @_;
 
+    $opts{'-compress'} //= $self->{'forcecompress'};
+
     require PDF::API2::Resource::XObject::Image::PNM;
-    my $obj = PDF::API2::Resource::XObject::Image::PNM->new($self->{'pdf'}, $file);
+    my $obj = PDF::API2::Resource::XObject::Image::PNM->new($self->{'pdf'}, $file, %opts);
 
     $self->{'pdf'}->out_obj($self->{'pages'});
 
@@ -2062,6 +2102,11 @@ sub image_pnm {
 
 Imports and returns a new PNG image object.  C<$file> may be either a filename or a filehandle.
 
+Note: PNG files that include an alpha (transparency) channel go through a
+relatively slow process of separating the transparency channel into a PDF SMask
+object.  Install PDF::API2::XS or Image::PNG::Libpng to speed this up by an
+order of magnitude.
+
 =cut
 
 sub image_png {
@@ -2387,13 +2432,20 @@ sub outlines {
     my $self = shift();
 
     require PDF::API2::Outlines;
-    $self->{'pdf'}->{'Root'}->{'Outlines'} ||= PDF::API2::Outlines->new($self);
-
     my $obj = $self->{'pdf'}->{'Root'}->{'Outlines'};
+    if ($obj) {
+        bless $obj, 'PDF::API2::Outlines';
+        $obj->{' api'} = $self;
+        weaken $obj->{' api'};
+    }
+    else {
+        $obj = PDF::API2::Outlines->new($self);
 
-    $self->{'pdf'}->new_obj($obj) unless $obj->is_obj($self->{'pdf'});
-    $self->{'pdf'}->out_obj($obj);
-    $self->{'pdf'}->out_obj($self->{'pdf'}->{'Root'});
+        $self->{'pdf'}->{'Root'}->{'Outlines'} = $obj;
+        $self->{'pdf'}->new_obj($obj) unless $obj->is_obj($self->{'pdf'});
+        $self->{'pdf'}->out_obj($obj);
+        $self->{'pdf'}->out_obj($self->{'pdf'}->{'Root'});
+    }
 
     return $obj;
 }
@@ -2458,7 +2510,8 @@ This module does not work with perl's -l command-line switch.
 PDF::API2 was originally written by Alfred Reibenschuh, extending code written
 by Martin Hosken.
 
-It is currently being maintained and developed by Steve Simms.
+It is currently being maintained and developed by Steve Simms, with patches from
+numerous contributors who are credited in the Changes file.
 
 =head1 LICENSE
 
